<?php

namespace Drupal\commerce_svea\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\Price;
use Drupal\commerce_svea\SveaManagerFactoryInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\profile\Entity\ProfileInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the Svea checkout payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "svea_checkout",
 *   label = "Svea Checkout",
 *   display_label = "Svea Checkout",
 *    forms = {
 *     "offsite-payment" = "Drupal\commerce_svea\PluginForm\OffsiteRedirect\SveaCheckoutForm",
 *   },
 *   requires_billing_information = FALSE,
 * )
 */
class SveaCheckout extends OffsitePaymentGatewayBase implements SveaCheckoutInterface {

  /**
   * The Svea manager factory.
   *
   * @var \Drupal\commerce_svea\SveaManagerFactoryInterface
   */
  protected $sveaManagerFactory;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Constructs a new SveaCheckout object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Drupal\commerce_svea\SveaManagerFactoryInterface $svea_manager_factory
   *   The Svea manager factory.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    PaymentTypeManager $payment_type_manager,
    PaymentMethodTypeManager $payment_method_type_manager,
    TimeInterface $time,
    SveaManagerFactoryInterface $svea_manager_factory,
    LoggerInterface $logger,
    ModuleHandlerInterface $module_handler
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);
    $this->sveaManagerFactory = $svea_manager_factory;
    $this->logger = $logger;
    $this->moduleHandler = $module_handler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('commerce_svea.manager_factory'),
      $container->get('logger.channel.commerce_svea'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'merchant_id' => '',
      'merchant_secret' => '',
      'purchase_country' => '',
      'locale' => 'sv-SE',
      'terms_path' => '',
      'update_billing_profile' => TRUE,
      'update_shipping_profile' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['merchant_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Merchant ID'),
      '#default_value' => $this->configuration['merchant_id'],
      '#required' => TRUE,
    ];
    $form['merchant_secret'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Merchant Secret'),
      '#default_value' => $this->configuration['merchant_secret'],
      '#required' => TRUE,
    ];
    $form['purchase_country'] = [
      '#type' => 'select',
      '#options' => [
        'SE' => $this->t('Sweden'),
        'NO' => $this->t('Norway'),
        'FI' => $this->t('Finland'),
        'DK' => $this->t('Denmark'),
        'DE' => $this->t('Germany'),
      ],
      '#title' => $this->t('Purchase country'),
      '#default_value' => $this->configuration['purchase_country'],
      '#required' => TRUE,
    ];
    $form['locale'] = [
      '#type' => 'select',
      '#title' => $this->t('Locale'),
      '#default_value' => $this->configuration['locale'],
      '#required' => TRUE,
      '#options' => [
        'sv-SE' => $this->t('Swedish (sv-SE)'),
        'nn-NO' => $this->t('Norwegian (nn-NO)'),
        'nb-NO' => $this->t('Norwegian (nb-NO)'),
        'fi-FI' => $this->t('Finnish (fi-FI)'),
        'da-DK' => $this->t('Danish (da-DK)'),
        'de-DE' => $this->t('German (de-DE)'),
      ],
    ];
    $form['terms_path'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Terms and conditions page path'),
      '#default_value' => $this->configuration['terms_path'],
      '#required' => TRUE,
    ];
    $form['update_billing_profile'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Update the billing customer profile with address information the customer enters at Svea.'),
      '#default_value' => $this->configuration['update_billing_profile'],
    ];
    $form['update_shipping_profile'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Update the shipping customer profile with address information the customer enters at Svea.'),
      '#default_value' => $this->configuration['update_shipping_profile'],
      '#access' => $this->moduleHandler->moduleExists('commerce_shipping'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['merchant_id'] = $values['merchant_id'];
      $this->configuration['merchant_secret'] = $values['merchant_secret'];
      $this->configuration['purchase_country'] = $values['purchase_country'];
      $this->configuration['locale'] = $values['locale'];
      $this->configuration['terms_path'] = $values['terms_path'];
      $this->configuration['update_billing_profile'] = $values['update_billing_profile'];
      $this->configuration['update_shipping_profile'] = $values['update_shipping_profile'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getNotifyUrl(): Url {
    $notify_url = parent::getNotifyUrl();
    $notify_url->setOption('query', [
      'checkout_order_id' => '{checkout.order.uri}',
    ]);
    return $notify_url;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $svea_order_id = $order->getData('svea_order_id');
    if (!$svea_order_id) {
      throw new PaymentGatewayException(sprintf('Cannot acknowledge the Svea order (Unknown Svea ID for order ID %s.', $order->id()));
    }

    /** @var \Drupal\commerce_payment\PaymentStorageInterface $payment_storage */
    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $payment = $payment_storage->loadByRemoteId($svea_order_id);

    // If a payment already exists in the system for the given order, it means
    // the order was already processed.
    if ($payment) {
      return;
    }
    try {
      $svea_manager = $this->sveaManagerFactory->get($this->getConfiguration());
      $svea_order = $svea_manager->getOrder($svea_order_id);

      if (isset($svea_order['EmailAddress'])) {
        $order->setEmail($svea_order['EmailAddress']);
      }

      // Update the billing profile if configured to do so.
      $configuration = $this->getConfiguration();
      if (!empty($configuration['update_billing_profile']) && isset($svea_order['BillingAddress'])) {
        $profile = $order->getBillingProfile();
        if (!$profile instanceof ProfileInterface) {
          /** @var \Drupal\profile\Entity\ProfileInterface $profile */
          $profile = $this->entityTypeManager->getStorage('profile')->create([
            'uid' => 0,
            'type' => 'customer',
          ]);
          $order->setBillingProfile($profile);
        }
        $this->populateProfile($profile, $svea_order['BillingAddress']);
        $profile->save();
      }

      // Update the shipping profile if configured to do so.
      if (!empty($configuration['update_shipping_profile']) && isset($svea_order['ShippingAddress'])) {
        $profiles = $order->collectProfiles();

        if (isset($profiles['shipping'])) {
          $this->populateProfile($profiles['shipping'], $svea_order['ShippingAddress']);
          $profiles['shipping']->save();
        }
      }

      // We have to save the order since the billing profile and/or the email
      // were potentially updated.
      $order->setRefreshState(OrderInterface::REFRESH_SKIP);
      $order->unlock();
      $order->save();
    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }

    // Create payment.
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $payment_storage->create([
      'state' => 'authorization',
      // Right for now, save order total as payment amount, because we don't
      // have any information from Svea for that value.
      'amount' => $order->getTotalPrice(),
      'payment_gateway' => isset($this->parentEntity) ? $this->parentEntity->id() : $this->entityId,
      'order_id' => $svea_order['ClientOrderNumber'],
      'test' => $this->getMode() === 'test',
      'remote_id' => $svea_order['OrderId'],
      'remote_state' => $svea_order['Status'],
    ]);

    // Capture the payment right a way.
    $this->capturePayment($payment);
  }

  /**
   * Populates the given profile with the given Svea address.
   *
   * @param \Drupal\profile\Entity\ProfileInterface $profile
   *   The profile to populate.
   * @param array $address
   *   The Svea address.
   */
  protected function populateProfile(ProfileInterface $profile, array $address) {
    $mapping = [
      // For B2B customers "fullName" field contains the company name.
      'FullName' => 'organization',
      'FirstName' => 'given_name',
      'LastName' => 'family_name',
      'CountryCode' => 'country_code',
      'PostalCode' => 'postal_code',
      'City' => 'locality',
      'StreetAddress' => 'address_line1',
      'StreetAddress2' => 'address_line2',
    ];
    foreach ($address as $key => $value) {
      if (isset($mapping[$key])) {
        $profile->address->{$mapping[$key]} = $value;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, Price $amount = NULL) {
    $this->assertPaymentState($payment, ['authorization']);

    // If not specified, capture the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $payment->setState('completed');
    $payment->setAmount($amount);
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment) {
    $this->assertPaymentState($payment, ['authorization']);

    try {
      $svea_manager = $this->sveaManagerFactory->get($this->getConfiguration());
      $svea_manager->cancelOrder($payment->getOrder());
      $payment->setState('voided');
      $payment->save();
    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function cancelOrder(OrderInterface $order) {
    try {
      $svea_manager = $this->sveaManagerFactory->get($this->getConfiguration());
      $svea_manager->cancelOrder($order);

      /** @var \Drupal\commerce_payment\PaymentStorageInterface $payment_storage */
      $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');

      // Void all the referenced payments after cancellation.
      $payments = $payment_storage->loadMultipleByOrder($order);
      foreach ($payments as $payment) {
        $payment->getState()->applyTransitionById('void');
        $payment->save();
      }
    }
    catch (\Exception $exception) {
      $this->logger->error($exception->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function fromMinorUnits($amount, string $currency_code): Price {
    $currency_storage = $this->entityTypeManager->getStorage('commerce_currency');
    /** @var \Drupal\commerce_price\Entity\CurrencyInterface $currency */
    $currency = $currency_storage->load($currency_code);
    $fraction_digits = $currency->getFractionDigits();

    if ($fraction_digits > 0) {
      $amount = Calculator::divide((string) $amount, pow(10, $fraction_digits), $fraction_digits);
    }

    return new Price((string) $amount, $currency_code);
  }

}
